$ifndef _MCP23S17_
$define _MCP23S17_
'
' Macros to Interface with a Microchip MCP23S17 16-bit SPI I/O peripheral
' For use with the Proton or Amicus18 BASIC compilers using an 18F25K20 device
'
    Include "Amicus18_SPI.inc"              ' Load the Amicus18 SPI macros into the program
'
' SPI interface lines
'
$ifndef MCP_CS
    $define MCP_CS PORTC.2                  ' MCP23S17 CS line
$endif

$ifndef MCP_RST
    $define MCP_RST PORTC.1                 ' MCP23S17 Reset Line (optional)
$endif

    Symbol cMicrochipAddress = $00          ' Default address of the MCP23S17 device
'
' MCP23S17 registers
'
    Symbol cIODIRA   = $00
    Symbol cIODIRB   = $01
    Symbol cIPOLA    = $02
    Symbol cIPOLB    = $03
    Symbol cGPINTENA = $04
    Symbol cGPINTENB = $05
    Symbol cDEFVALA  = $06
    Symbol cDEFVALB  = $07
    Symbol cINTCONA  = $08
    Symbol cINTCONB  = $09
    Symbol cIOCONA   = $0A
    Symbol cIOCONB   = $0B
    Symbol cGPPUA    = $0C
    Symbol cGPPUB    = $0D
    Symbol cINTFA    = $0E
    Symbol cINTFB    = $0F
    Symbol cINTCAPA  = $10
    Symbol cINTCAPB  = $11
    Symbol cGPIOA    = $12
    Symbol cGPIOB    = $13
    Symbol cOLATA    = $14
    Symbol cOLATB    = $15
'
' Control byte for SPI operations [0100][A2][A1][A0][R=1/W=0]
'
    Symbol cControlByteMaskRead = %01000001 | (cMicrochipAddress << 1)
    Symbol cControlByteMaskWrite = %01000000 | (cMicrochipAddress << 1)

    Symbol cControl_IOCON = %00011000        ' Init value for IOCON - bits details in following comment
(*
 Register IOCON = [BANK][MIRROR][SEQOP][DISSLW][HAEN][ODR][INTPOL][-Unused]

 Bit-7 BANK:    Controls how the registers are addressed
                    1 = The registers associated With each Port are separated into different banks
                    0 = The registers are in the same bank (addresses are sequential)
 Bit-6 MIRROR:  INT Pins Mirror Bit
                    1 = The INT pins are internally connected
                    0 = The INT pins are Not connected. INTA is associated With PortA And INTB is associated With PortB
 Bit-5 SEQOP:   Sequential Operation mode Bit.
                    1 = Sequential operation disabled, address pointer does Not increment.
                    0 = Sequential operation Enabled, address pointer increments.
 Bit-4 DISSLW:  Slew Rate control Bit For SDA Output.
                    1 = Slew rate disabled.
                    0 = Slew rate Enabled.
 Bit-3 HAEN:    Hardware Address Enable Bit (MCP23S17 only).
                    Address pins are always Enabled on MCP23017.
                    1 = Enables the MCP23S17 address pins.
                    0 = Disables the MCP23S17 address pins.
 Bit-2 ODR:     This Bit configures the INT pin As an open-drain Output.
                    1 = Open-drain Output (overrides the INTPOL Bit).
                    0 = Active driver Output (INTPOL Bit sets the polarity).
 Bit-1 INTPOL:  This Bit sets the polarity of the INT Output pin.
                    1 = Active-High.
                    0 = Active-Low.
 Bit-0 Unimplemented: Read As 0.
*)
'
' Variables used to cache the state of Direction and Port registers
'
    Dim bShadowDirRegA As Byte
    Dim bShadowDirRegB As Byte
    Dim bShadowDataRegA As Byte
    Dim bShadowDataRegB As Byte
'
' Variables used for passing parameters to the macros (better left as System types in bankless RAM)
'
    Dim bRegAddress As Byte System
    Dim bDataByte As Byte System
    Dim bMCPtemp As bDataByte

'----------------------------------------------------------------------------
    GoTo MCP23S17_Main                                  ' Jump over the subroutines

'****************************************************************************
'* Name    : Write_MCP                                                      *
'* Puspose : Write to one of the MCP23S17 registers                         *
'*                                                                          *
'* Input   : bRegAddress holds the address to write                         *
'*         : bDataByte holds the data to write to the register              *
'* Output  : None                                                           *
'* Notes   : None                                                           *
'****************************************************************************
$define MCP23S17_Write(pRegAddress, pDataByte) '
    bDataByte = pDataByte                      '
    WREG = pRegAddress : _Write_MCP

_Write_MCP Macro-
    GoSub __Write_MCP
Endm

#ifMacro- _Write_MCP, _WritePinPortA, _WritePinPortB, _WritePinTrisA, _WritePinTrisB
__Write_MCP:
    bRegAddress = WREG
    Clear MCP_CS
    DelayCS 1
    SPI_Write(cControlByteMaskWrite)
    SPI_Write(bRegAddress)
    SPI_Write(bDataByte)
    Set MCP_CS
    Return
#endIfMacro-

'****************************************************************************
'* Name    : Read_MCP                                                       *
'* Puspose : Read from one of the MCP23S17 registers                        *
'                                                                           *
'* Input   : bRegAddress holds the adddress To write                        *
'* Output  : SSPBUF holds the byte read from the MCP device                 *
'* Notes   : Uses  the second parameter to return the value                 *
'****************************************************************************
$define MCP23S17_Read(pRegAddress, pReturn) '
    WREG = pRegAddress '
    _Read_MCP          '
    pReturn = WREG

_Read_MCP Macro-
    GoSub __Read_MCP
Endm

#ifMacro- _Read_MCP, _ReadPortA, _ReadPortB
__Read_MCP:
    bRegAddress = WREG
    Clear MCP_CS
    DelayCS 1
    SPI_Write(cControlByteMaskRead)
    SPI_Write(bRegAddress)
    WREG = SPI_Read()
    Set MCP_CS
    Return
#endIfMacro-

'****************************************************************************
'* Name    : ReadPortA                                                      *
'* Purpose : Reads PortA                                                    *
'* Notes   : Saves the port value to a shadow variable                      *
'****************************************************************************
$define MCP23S17_ReadPortA() _ReadPortA

_ReadPortA Macro- \Byte
    'MCP23S17_Read(cGPIOA,WREG)
    WREG = cGPIOA
    GoSub __Read_MCP
    bShadowDataRegA = WREG
    Return_Wreg
Endm
'****************************************************************************
'* Name    : ReadPortB                                                      *
'* Purpose : Reads PortB                                                    *
'* Notes   : Saves the port value to a shadow variable                      *
'****************************************************************************
$define MCP23S17_ReadPortB() _ReadPortB

_ReadPortB Macro- \Byte
    'MCP23S17_Read(cGPIOB,WREG)
    WREG = cGPIOB
    GoSub __Read_MCP
    bShadowDataRegB = WREG
    Return_Wreg
Endm

'****************************************************************************
'* Name    : WritePortA                                                     *
'* Purpose : Writes PortA Latch register with pDataByte value               *
'*                                                                          *
'* Notes   : Saves the port value to a shadow variable                      *
'****************************************************************************
$define MCP23S17_WritePortA(pDataByte) '
    bShadowDataRegA = pDataByte        '
    MCP23S17_Write(cOLATA, pDataByte)

'****************************************************************************
'* Name    : WritePortB                                                     *
'* Purpose : Writes PortB Latch register with pDataByte value               *
'*                                                                          *
'* Notes   : Saves the port value to a shadow variable                      *
'****************************************************************************
$define MCP23S17_WritePortB(pDataByte) '
    bShadowDataRegB = pDataByte        '
    MCP23S17_Write(cOLATB, pDataByte)

'****************************************************************************
'* Name    : SetTrisPortA                                                   *
'* Purpose : Sets PortA pin's direction 0=Output 1=Input                    *
'*                                                                          *
'* Notes   : Saves the tris value to a shadow variable                      *
'****************************************************************************
$define MCP23S17_SetTrisPortA(pDirection) '
    MCP23S17_Write(cIODIRA, pDirection)   '
    bShadowDirRegA = pDirection

'****************************************************************************
'* Name    : SetTrisPortB
'* Purpose : Sets PortB pin's direction 0=Output 1=Input
'*
'* Notes   : Saves the tris value to a shadow variable
'****************************************************************************
$define MCP23S17_SetTrisPortB(pDirection) '
    MCP23S17_Write(cIODIRB, pDirection)   '
    bShadowDirRegB = pDirection

'****************************************************************************
'* Name    : WritePinPortA                                                  *
'* Purpose : Write to a bit of PortA                                        *
'* Parameters : pPin Output pin to set or clear                             *
'*              pHiLo 1 = high, 0 = low                                     *
'* Notes   : Uses the shadow port to read the previous state                *
'****************************************************************************
$define MCP23S17_WritePinPortA(pPin,pHiLo) '
    bMCPtemp = pHiLo '
    WREG = pPin      '
    _WritePinPortA

_WritePinPortA Macro-
    GoSub __WritePinPortA
Endm

#ifMacro- _WritePinPortA
__WritePinPortA:
    PRODL = WREG
    PRODH =  1 << PRODL
    If bMCPtemp = 1 Then
        PRODH = PRODH | bShadowDataRegA
    Else
        PRODH = ~PRODH
        PRODH = PRODH & bShadowDataRegA
    EndIf
    MCP23S17_Write(cOLATA, PRODH)
    Return
#endIfMacro

'****************************************************************************
'* Name    : WritePinPortB                                                  *
'* Purpose : Write to a bit of PortB                                        *
'* Parameters : pPin Output pin to set or clear                             *
'*              pHiLo 1 = high, 0 = low                                     *
'* Notes   : Uses the shadow port to read the previous state                *
'****************************************************************************
$define MCP23S17_WritePinPortB(pPin,pHiLo) '
    bMCPtemp = pHiLo    '
    WREG = pPin         '
    _WritePinPortB

_WritePinPortB Macro-
    GoSub __WritePinPortB
Endm

#ifMacro- _WritePinPortB
__WritePinPortB:
    PRODL = WREG
    PRODH =  1 << PRODL
    If bMCPtemp = 1 Then
        PRODH = PRODH | bShadowDataRegB
    Else
        PRODH = ~PRODH
        PRODH = PRODH & bShadowDataRegB
    EndIf
    MCP23S17_Write(cOLATB, PRODH)
    Return
#endIfMacro-

'****************************************************************************
'* Name    : WritePinTrisA                                                  *
'* Purpose : Sets or clears the Tris of a Single Pin from PortA pins        *
'* Parameters : pPin Tris pin to set or clear (0 - 7)                       *
'*              pDirection 1 = Input, 0 = Output                            *
'****************************************************************************
$define MCP23S17_WritePinTrisA(pPin,pDirection) '
    bMCPtemp = pDirection   '
    WREG = pPin             '
    _WritePinTrisA

_WritePinTrisA Macro-
    GoSub __WritePinTrisA
Endm

#ifMacro- _WritePinTrisA
__WritePinTrisA:
    PRODL = WREG
    PRODH =  1 << PRODL
    If bMCPtemp = 1 Then
        PRODH = PRODH | bShadowDirRegA
    Else
        PRODH = ~PRODH
        PRODH = PRODH & bShadowDirRegA
    EndIf
    bShadowDirRegA = PRODH
    MCP23S17_Write(cIODIRA, PRODH)
    Return
#endIfMacro
'****************************************************************************
'* Name    : WritePinTrisB                                                  *
'* Purpose : Sets or clears the Tris of a Single Pin from PortB pins        *
'* Parameters : pPin Tris pin to set or clear (0 - 7)                       *
'*              pDirection 1 = Input, 0 = Output                            *
'****************************************************************************
$define MCP23S17_WritePinTrisB(pPin,pDirection) '
    bMCPtemp = pDirection   '
    WREG = pPin             '
    _WritePinTrisB

_WritePinTrisB Macro-
    GoSub __WritePinTrisB
Endm

#ifdef _WritePinTrisB#Req
__WritePinTrisB:
    PRODL = WREG
    PRODH =  1 << PRODL
    If bMCPtemp = 1 Then
        PRODH = PRODH | bShadowDirRegB
    Else
        PRODH = ~PRODH
        PRODH = PRODH & bShadowDirRegB
    EndIf
    bShadowDirRegB = PRODH
    MCP23S17_Write(cIODIRB, PRODH)
    Return
#endif

'****************************************************************************
'* Name    : ReadTrisPortA                                                  *
'* Purpose : Gets PortA pins direction 0=Output 1=Input                     *
'* Notes   : Reads the shadow register instead of reading the register      *
'****************************************************************************
$define MCP23S17_ReadTrisPortA() bShadowDirRegA

'****************************************************************************
'* Name    : ReadTrisPortB                                                  *
'* Purpose : Gets PortB pins direction 0=Output 1=Input                     *
'* Notes   : Reads the shadow register instead of reading the register      *
'****************************************************************************
$define MCP23S17_ReadTrisPortB() bShadowDirRegB

'****************************************************************************
'* Name    : SetPullUpsPortA                                                *
'* Purpose : Set the Pullups resistors  for any Pin of PortA                *
'* Notes   : Pullups can be enabled/disabled independently                  *
'*           for any pin of the port 0=Disabled  1=Enabled                  *
'****************************************************************************
$define MCP23S17_SetPullUpsPortA(pPullUps) MCP23S17_Write(cGPPUA, pPullUps)

'****************************************************************************
'* Name    : SetPullUpsPortB                                                *
'* Purpose : Sets the Pullups resistors  for any Pin of PortB               *
'* Notes   : Pullups can be enabled/disabled independently                  *
'*           for any pin of the port 0=Disabled  1=Enabled                  *
'****************************************************************************
$define MCP23S17_SetPullUpsPortB(pPullUps) MCP23S17_Write(cGPPUB, pPullUps)

'****************************************************************************
'* Initialise the MCP23S17                                                  *
'* Chip_Address must be forced to "000"                                     *
'* regardless of actual Mcpchip_Address value.                              *
'* After initialisation, once IOCON.HAEN is enabled,                        *
'* Mcpchip_Address is taken into account by the chip.                       *
'****************************************************************************
$define MCP23S17_Init() _MCP23S17_Init

_MCP23S17_Init Macro-
    GoSub __MCP23S17_Init
    Endm
 
#ifMacro- _MCP23S17_Init   
__MCP23S17_Init:
    High MCP_CS
$ifdef MCP_RST                                   ' Is a reset line being used?
    Low MCP_RST                                  ' Yes. So pull the reset line low
    DelayMS 10                                   ' Wait 10 milliseconds
    Set MCP_RST                                  ' Bring the device out of reset
$endif

    SPI_Open(SPI_FOSC_16, MODE_00, SMPEND)        ' Open the SPI peripheral
    Clear MCP_CS
    DelayCS 1
    SPI_Write(cControlByteMaskWrite & %11110001)  ' Force ChipAdd to 000
    SPI_Write(cIOCONA)                            ' Writing to cIOCONA writes also to cIOCONB
    SPI_Write(cControl_IOCON)
    Set MCP_CS
    bShadowDirRegA = %11111111
    bShadowDirRegB = %11111111
    bShadowDataRegA = %00000000
    bShadowDataRegB = %00000000
    DelayMS 10
    Return
#endIfMacro-

'----------------------------------------------------------------------------------------
MCP23S17_Main:
    MCP23S17_Init()
        
$endif



